/*
 * IGS vga chip register tool
 *
 */
#include <linux/init.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/miscdevice.h>
#include <linux/delay.h>
#include <linux/stat.h>
#include <linux/proc_fs.h>
#include <linux/config.h>

#include <asm/hardware.h>
#include <asm/uaccess.h>

/*****************************************************************************/
#include "igsvga.h"

#define CRT_CMD 1
#define ATR_CMD 2
#define SEQ_CMD 3
#define GRF_CMD 4
#define MISC_CMD 5
#define EXT_CMD 6
#define PAL_CMD 7
#define	TV_CMD	8

#define	SEQ_I	0x3C4
#define	SEQ_D	0x3C5

#define	CRT_I	0x3D4
#define	CRT_D	0x3D5


static int open_igsvga(struct inode* inodep, struct file* filep);
static int release_igsvga(struct inode* inodep, struct file* filep);
static int igsvga_ioctl(struct inode* inodep, struct file* filep, unsigned int cmd, unsigned long arg);
void igs_setregs(int regtype, char* regs, int n);
void igs_getregs(int regtype, char* regs, int n);


#define	IGSVGA_MAJOR	100
static const char igsvga_name[] = "igsvga";
static volatile char* pVidPort;



static struct file_operations igsvga_fops = {
	NULL,			/* lseek              */
	NULL,			/* read               */
	NULL,			/* write              */
	NULL,			/* no special readdir            */
	NULL,			/* no special select             */
	igsvga_ioctl,
	NULL,			/* no special mmap               */
	open_igsvga,
	NULL,			/* flush */
	release_igsvga,
	NULL,			/* no special fsync              */
	NULL,			/* no special fasync             */
	NULL,			/* no special check_media_change */
	NULL			/* no special revaldate          */
};


void igs_getregs(int regtype, char* regs, int n)
{
	int i;
	volatile char c;

	for (i=0; i<n; i++)
	{
		switch(regtype)
		{
			case CRT_CMD:
				pVidPort[0x3d4] = i;
				udelay(10);
				regs[i] = pVidPort[0x3d5];
				break;
			case ATR_CMD:
				c = pVidPort[0x3da];
				udelay(10);
				pVidPort[0x3c0] = i;
				udelay(10);
				regs[i] = pVidPort[0x3c1];
				break;
			case SEQ_CMD:
				pVidPort[0x3c4] = i+1;
				udelay(10);
				regs[i] = pVidPort[0x3c5];
				break;
			case GRF_CMD:
			case EXT_CMD:
				pVidPort[0x3ce] = i;
				udelay(10);
				regs[i] = pVidPort[0x3cf];
				break;
			case MISC_CMD:
				regs[i] = pVidPort[0x3cc];
				break;
/*
			case PAL_CMD:
				pVidPort[0x3ce] = 0x56;
				c = pVidPort[0x3cf];
				pVidPort[0x3cf] = c | 4;
				regs[i] = pVidPort[0x3c6];
				pVidPort[0x3cf] = c;
				break;
*/
			case TV_CMD:
				regs[i] = pVidPort[0x3cc];
				break;

		}
	}
}

void igs_setregs(int regtype, char* regs, int n)
{
	int i;
	volatile char c;

	for (i=0; i<n; i++)
	{
		switch(regtype)
		{
			case CRT_CMD:
				pVidPort[0x3d4] = i;
				udelay(10);
				pVidPort[0x3d5] = regs[i];
				break;
			case ATR_CMD:
				c = pVidPort[0x3da];
				udelay(10);
				pVidPort[0x3c0] = i;
				udelay(10);
				pVidPort[0x3c0] = regs[i];
				break;
			case SEQ_CMD:
#if 0
				if (!i)
				{
					pVidPort[SEQ_I] = 0x01;
			    		pVidPort[SEQ_D] = regs[i] | 0x20;
				}
#endif
				pVidPort[0x3c4] = i+1;
				udelay(10);
				pVidPort[0x3c5] = regs[i];
				break;


			case GRF_CMD:
			case EXT_CMD:
				pVidPort[0x3ce] = i;
				udelay(10);
				pVidPort[0x3cf] = regs[i];
				break;
			case MISC_CMD:
				pVidPort[0x3c2] = regs[0];
				break;
/*
			case PAL_CMD:
				pVidPort[0x3ce] = 0x56;
				c = pVidPort[0x3cf];
				pVidPort[0x3cf] = c | 4;
				pVidPort[0x3c6] = regs[i];
				pVidPort[0x3cf] = c;
				break;
*/
		}
	}
}

	
static int open_igsvga(struct inode* inodep, struct file* filep)
{
	MOD_INC_USE_COUNT;

//	pVidPort = (volatile char*)0xE8800000;		//2.3
	pVidPort = (volatile char*)0xE1800000;		//2.2

	return 0;
}


static int release_igsvga(struct inode* inodep, struct file* filep)
{
	MOD_DEC_USE_COUNT;

	return 0;
}


static int igsvga_ioctl(struct inode* inodep, struct file* filep, unsigned int cmd, unsigned long arg)
{
long unsigned flags;
IGSREGS igs;
int	pArray;
volatile char c;

//	printk("igsvga_ioctl: cmd = 0x%X, arg = 0x%X.\n",cmd, *(int*)arg);

        pArray = *(int *) arg;		//get the address of user IGSREGS

	switch (cmd)
	{
        case IGSVGA_WRITE:

//            if (verify_area(VERIFY_READ, (void*)pArray, sizeof(IGSREGS) == -EFAULT))
//		return(-EFAULT);

//            memcpy_fromfs(&igs, (void*)pArray, sizeof(IGSREGS)); //get new regs
		copy_from_user(&igs, (void *)pArray, sizeof(IGSREGS));

		save_flags(flags);
		cli();

//		c = pVidPort[0x3da];

	        /* synchronous reset on */
		pVidPort[SEQ_I] = 0x00;
		pVidPort[SEQ_D] = 0x01;

		igs_setregs(SEQ_CMD,(char*)&igs.seq_reg,4);
		igs_setregs(MISC_CMD,(char*)&igs.misc_reg,1);

	        /* synchronous reset off */
	        pVidPort[SEQ_I] = 0x00;
		pVidPort[SEQ_D] = 0x03;


		/* Unprotect CRT regs 0..7 */
		pVidPort[CRT_I] = 0x11;
		pVidPort[CRT_D] = igs.crt_reg[0x11] & 0x7F;
		igs_setregs(CRT_CMD,(char*)&igs.crt_reg,25);

		igs_setregs(ATR_CMD,(char*)&igs.atr_reg,20);

		igs_setregs(GRF_CMD,(char*)&igs.grf_reg,256);

		pVidPort[0x3ce] = 0xB9;		//toggle this port to latch
		udelay(10);
		pVidPort[0x3cf] = 0x80;
		udelay(10);
		pVidPort[0x3cf] = 0x0;

		// turn video on
		c = pVidPort[0x3da];
		udelay(10);
		pVidPort[0x3c0] = 0x20;

		restore_flags(flags);

//	    printk("igsvga: Wrote IGS regs...\n");
	    break;

        case IGSVGA_READ:

		//first read all current values...
		save_flags(flags);
		cli();

		igs_getregs(GRF_CMD,(char*)&igs.grf_reg,256);
		igs_getregs(CRT_CMD,(char*)&igs.crt_reg,25);
//		c = pVidPort[0x3da];
		igs_getregs(ATR_CMD,(char*)&igs.atr_reg,20);
		igs_getregs(SEQ_CMD,(char*)&igs.seq_reg,4);
		igs_getregs(MISC_CMD,(char*)&igs.misc_reg,1);

		// turn video on
		c = pVidPort[0x3da];
		udelay(10);
		pVidPort[0x3c0] = 0x20;

		restore_flags(flags);

		copy_to_user((void *)pArray, &igs, sizeof(IGSREGS));
//		if (verify_area(VERIFY_WRITE, (void*)pArray, sizeof(IGSREGS) == -EFAULT))
//			return(-EFAULT);

//		memcpy_tofs((void*)pArray, &igs, sizeof(IGSREGS));
//		printk("igsvga: Returning IGS regs...\n");

		break;


	default:
		return -EINVAL;
	}

	return 0;
}


#ifndef MODULE
long igsvga_init(long mem_start, long mem_emd)
{
	if (igsvga_major = register_chrdev(0, igsvga_name, &igsvga_fops))
		printk("Unable to get major for igsvga device\n");
	else
		printk("igsvga V.0.0.2 for Armlinux 2.3: major = %d... Woody 1998.\n", IGSVGA_MAJOR);

	return mem_start;
}
#else
int init_module(void)
{
int	temp;

//	printk("IGSVGA: init_module().\n");
//	nwprintk("IGSVGA: init_module().\n");
	if ((temp = register_chrdev(IGSVGA_MAJOR, igsvga_name, &igsvga_fops)) == -EBUSY)
	{
		printk("Unable to get major for igsvga device (%d)\n",temp);
		return -EIO;
	}
	else
	{
		printk("igsvga V.0.0.2 for Linux 2.3: major = %d... Woody 1998.\n", IGSVGA_MAJOR);
//		nwprintk("igsvga V.0.0.2: major = %d... Woody 1998.\n", IGSVGA_MAJOR);
	}
	return 0;
}

void cleanup_module(void)
{
	unregister_chrdev(IGSVGA_MAJOR, igsvga_name);
}
#endif
